package ru.dbotthepony.mc.otm.block.entity.matter

import net.minecraft.core.BlockPos
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.ForgeConfigSpec
import ru.dbotthepony.mc.otm.config.ConciseBalanceValues
import ru.dbotthepony.mc.otm.block.entity.MatteryWorkerBlockEntity
import ru.dbotthepony.mc.otm.capability.energy.BlockEnergyStorageImpl
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.energy.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matter.IPatternState
import ru.dbotthepony.mc.otm.capability.matter.PatternState
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.HandlerFilter
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.DecimalConfigValue
import ru.dbotthepony.mc.otm.core.math.defineDecimal
import ru.dbotthepony.mc.otm.graph.Graph6Node
import ru.dbotthepony.mc.otm.graph.matter.MatterGraph
import ru.dbotthepony.mc.otm.menu.matter.MatterScannerMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities
import ru.dbotthepony.mc.otm.registry.MNames
import ru.dbotthepony.mc.otm.core.util.WriteOnce
import ru.dbotthepony.mc.otm.graph.matter.MatterNode
import ru.dbotthepony.mc.otm.matter.MatterManager
import java.util.*
import kotlin.math.pow

class MatterScannerBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) :
	MatteryWorkerBlockEntity<MatteryWorkerBlockEntity.ItemJob>(MBlockEntities.MATTER_SCANNER, p_155229_, p_155230_, ::ItemJob) {

	val container = MatteryContainer(this::itemContainerUpdated, 1).also(::addDroppableContainer)
	val energy = WorkerEnergyStorage(this::powerLevelUpdated, ENERGY_VALUES)
	val itemHandler = container.handler(object : HandlerFilter {
		override fun canInsert(slot: Int, stack: ItemStack): Boolean {
			return MatterManager.canDecompose(stack)
		}

		override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
			return isIdling
		}
	})

	val matterNode = object : MatterNode() {
		override fun onPatternAdded(state: IPatternState) {
			if (idleReason == IdleReason.OBSERVING) {
				isIdling = false
			}
		}

		override fun onPatternRemoved(state: IPatternState) {
			if (idleReason == IdleReason.OBSERVING) {
				isIdling = false
			}
		}

		override fun onPatternUpdated(newState: IPatternState, oldState: IPatternState) {
			if (idleReason == IdleReason.OBSERVING) {
				isIdling = false
			}
		}
	}

	init {
		exposeItemsGlobally(itemHandler)
		exposeEnergyGlobally(energy)
		exposeGlobally(MatteryCapability.MATTER_NODE, matterNode)

		savetable(::container, INVENTORY_KEY)
		savetable(::energy, ENERGY_KEY)
	}

	override fun invalidateCaps() {
		super.invalidateCaps()
		matterNode.isValid = false
	}

	override fun setRemoved() {
		super.setRemoved()
		matterNode.isValid = false
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return MatterScannerMenu(containerID, inventory, this)
	}

	override fun onJobFinish(job: ItemJob): Status {
		val stack = job.itemStack
		if (stack.isEmpty || !MatterManager.hasMatterValue(stack)) return Status.SUCCESS

		var findState: IPatternState? = null

		for (state in matterNode.graph.patterns.filter { it.item === stack.item }) {
			if (findState == null && state.researchPercent < 1.0) {
				findState = state
			} else if (findState != null && findState.researchPercent < state.researchPercent && state.researchPercent < 1.0) {
				findState = state
			}
		}

		val researchAdvance = MatterManager.getResearchAdvance(stack.item)

		val new: IPatternState =
			if (findState != null) {
				PatternState(findState.id, stack.item, (findState.researchPercent + researchAdvance).coerceAtMost(1.0))
			} else {
				PatternState(UUID.randomUUID(), stack.item, researchAdvance)
			}

		if (!matterNode.graph.insertPattern(new, onlyUpdate = false, simulate = false).isFailed) {
			return Status.SUCCESS
		} else {
			return Status.FAILURE_WAIT
		}
	}

	override fun computeNextJob(): Pair<ItemJob?, IdleReason?> {
		if (energy.batteryLevel.isZero) {
			return null to IdleReason.POWER
		}

		val stack = container.getItem(0)
		if (stack.isEmpty || !MatterManager.canDecompose(stack)) return null to IdleReason.ITEM

		var findState: IPatternState? = null

		for (state in matterNode.graph.patterns.filter { it.item === stack.item }) {
			if (state.researchPercent < 1.0) {
				findState = state
			} else if (state.researchPercent >= 1.0) {
				return null to IdleReason.ITEM
			} else if (findState != null && findState.researchPercent < state.researchPercent) {
				findState = state
			}
		}

		val researchAdvance = MatterManager.getResearchAdvance(stack.item)

		val new: IPatternState =
			if (findState != null) {
				PatternState(findState.id, stack.item, (findState.researchPercent + researchAdvance).coerceAtMost(1.0))
			} else {
				PatternState(UUID.randomUUID(), stack.item, researchAdvance)
			}

		if (!matterNode.graph.insertPattern(new, onlyUpdate = false, simulate = true).isFailed) {
			val copy = stack.copy().also { it.count = 1 }
			stack.shrink(1)
			container.setChanged()
			val complexity = MatterManager.get(copy).complexity
			return ItemJob(copy, (if (complexity > 1.0) complexity.pow(1.25) else complexity.pow(0.5)), BASE_CONSUMPTION) to null
		}

		return null to IdleReason.ITEM
	}

	override fun setLevel(level: Level) {
		super.setLevel(level)
		matterNode.discover(this)
	}

	companion object {
		private val BASE_CONSUMPTION get() = _BASE_CONSUMPTION.get()

		private var _BASE_CONSUMPTION: DecimalConfigValue by WriteOnce()
		var ENERGY_VALUES: ConciseBalanceValues by WriteOnce()
			private set

		fun registerConfig(builder: ForgeConfigSpec.Builder) {
			builder.push(MNames.MATTER_SCANNER)

			ENERGY_VALUES = BlockEnergyStorageImpl.makeConfigEntry(builder, capacity = Decimal(80_000), throughput = Decimal(400))
			_BASE_CONSUMPTION = builder.defineDecimal("baseConsumption", Decimal(40), Decimal.ONE)

			builder.pop()
		}
	}
}
